/* Program to send data to bpp device */

#include "version.h"

#include <sys/types.h>
#include <fcntl.h>
#include <sys/ioccom.h>
#include <sbusdev/bpp_io.h>

#include <stdio.h>
#include <sys/uio.h>
#include <signal.h>

#ifndef BUFSIZE
#define BUFSIZE 1024			/* Buffer size */
#endif  /* BUFSIZE */
#define BPPDEV "/dev/bpp0"		/* Output device */

#define MIN_WAIT_INTERVAL  100		/* in microsec */
#define MAX_WAIT_INTERVAL  15*1000*1000	/* in microsec */
#define RESET_DURATION 1000		/* in microsec */

#define REPORT_BYTES 1024
  /* In verbose mode, report that this number of bytes has been written */

/* Declarations */
struct buffer{
  int start, stop;    /* Indeces to start and stop positions in the buffer.
			 buf[start] to buf[stop-1] are valid.  */
  char buf[BUFSIZE];
};

/* Global variables */
unsigned long cumcnt=0;
int outfd;
FILE *infp;
int verbose=0;


/*** Main ***/
main(int argc, char *argv[]){

  extern int outfd;
  extern FILE *infp;
  extern int verbose;

  char *bppdev, *filename;
  struct bpp_transfer_parms bpppar;
  int j, do_reset=0;

  extern void sig_proc();  /* signal handler */

  /* Command line parser.  If more than one files are given,
     only first one is valid.  The file name cannot begin with '-'.*/
  filename=NULL;   /* If filename remains NULL, no file name is given
		      as an argument.*/
  for(j=argc-1; j>=1 ; j--){
    if( argv[j][0] == '\000' ){
      /* Ignore null string */

    }else if( argv[j][0] == '-' ){
      if( strcmp(argv[j],"-help")==0 ) exit(bpp_help());
      if( strcmp(argv[j],"-verb")==0 ) verbose=1;
      if( strcmp(argv[j],"-reset")==0 ) do_reset=1;
    }else{
      /* File name */
      filename=argv[j];
    }
  }

  if (verbose)  fprintf(stderr,"\t*** %s %s ***\n\t    %s\t%s\n\n",
			PROG_NAME,VERSION, DATE, AUTHOR);

  bppdev=BPPDEV;

  outfd = open(bppdev, O_WRONLY);
  if (outfd == 0){
    fprintf(stderr,"bppout: Failed to open %s\n",bppdev);
      exit(1);
    }

  /* Set handsake mode to BPP_BUSY_HS rather than BPP_ACK_HS.  If the
   * defalut mode of BPP_ACK_HS were used, write() might behave
   * unexpectedly, i.e. it would sometimes returns more than the number
   * of requested bytes.  Though I don't know the reason, this problem
   * can be fixed by setting the handshake mode to BPP_BUSY_HS.
   */
  ioctl(outfd, BPPIOC_GETPARMS, &bpppar);
  bpppar.write_handshake=BPP_BUSY_HS;
  ioctl(outfd, BPPIOC_SETPARMS, &bpppar);

  if(verbose){
    fprintf(stderr, " HNDSHK=%d SETUPT=%d STRWDT=%d TIMOUT=%d\n",
	    bpppar.write_handshake, bpppar.write_setup_time,
	    bpppar.write_strobe_width, bpppar.write_timeout);
  }

  if(do_reset) exit(bpp_reset(outfd));

  if(filename == NULL){
    infp = stdin;
    if(verbose) fprintf(stderr,"... Reading data from the standard input.\n");
   }else{
    infp=fopen(filename,"r");
    if( infp==NULL ){
      fprintf(stderr,"bppout: Failed to open %s\n",filename);
      exit(1);
    }
    if(verbose) fprintf(stderr,"... Reading %s.\n",filename);
  }

  /* Catch signals */
  signal(SIGINT,  sig_proc);
  signal(SIGQUIT, sig_proc);
  signal(SIGTERM, sig_proc);
  signal(SIGTSTP, sig_proc);

  /* Now ready to process. */
  bpp_out();

  close(bppdev);
  fclose(infp);
  exit(0);
}


int bpp_out(){

  struct buffer b;			/* b is a real object. */
  int j;
  int cnt;
  extern unsigned long cumcnt;		/* To be exported to signal handler. */
  extern int outfd;
  extern FILE *infp;
  extern int verbose;

  wait_ready();				/* Wait for the port to be writable. */

  /* Repeat until input ends. */
  while( fill_buf(infp, &b) > 0){

    /* Repeat until start pointer reaches stop pointer. */
    while( b.start < b.stop){

      cnt=write(outfd, &b.buf[b.start], (int) (b.stop-b.start));

      if ( cnt<=0 ){
	wait_ready();

      }else if(cnt<BUFSIZE){
	b.start += cnt;
	cumcnt += cnt;
	if (verbose  &&  cumcnt%REPORT_BYTES < cnt )
	  fprintf(stderr, "... %d bytes written\n", cumcnt);

      }else if(cnt==BUFSIZE){
	b.start += cnt;
	cumcnt += cnt;
	if (verbose  &&  cumcnt%REPORT_BYTES < cnt )
	  fprintf(stderr, "... %d bytes\n", cumcnt);
 	break;

      }else{
#ifdef DEBUG
	fprintf(stderr, "* %d %d  %d %d \n",(int)cumcnt, cnt, b.start, b.stop);
	for (j=0; j<cnt; j++)  fprintf(stderr, "  %d", b.buf[j]  );
	fprintf(stderr, "\n");
	b.start += cnt;
	cumcnt += cnt;
#else
	fprintf(stderr, "bppout: Error in write() system call\n");
	exit(1);
#endif
      }

    }

  }

  if(verbose)
    fprintf(stderr, "### %d bytes written in tolal.\n",(int) cumcnt);

  return 0;

}

int wait_ready(){
  extern int outfd;
  extern FILE *infp;
  extern int verbose;
  unsigned usec;
  static struct bpp_error_status bpp_err[1];

  if(verbose  &&  ioctl(outfd, BPPIOC_TESTIO) !=0){
    ioctl(outfd, BPPIOC_GETERR, bpp_err);
    report_err_stat(bpp_err);
  }

  usec= MIN_WAIT_INTERVAL;  /* Begin with MIN_WAIT_INTERVAL microsec. */

  while( ioctl(outfd, BPPIOC_TESTIO) !=0 ){
    usleep(usec);
    usec =  (usec*2<MAX_WAIT_INTERVAL) ? usec*2 : MAX_WAIT_INTERVAL;
  }
  return 0;
}

int report_err_stat(struct bpp_error_status *bpp_err){
  int pin_stat;

  if( bpp_err->timeout_occurred )
    fprintf(stderr, "*** Timeout occured\n");
  if( bpp_err->bus_error )
    fprintf(stderr, "*** SBus error occured\n");

  if( pin_stat = bpp_err->pin_status ){
    if(pin_stat&BPP_ERR_ERR)   fprintf(stderr,"*** Error pin active\n");
    if(pin_stat&BPP_SLCT_ERR)  fprintf(stderr,"*** Select pin active\n");
    if(pin_stat&BPP_PE_ERR)    fprintf(stderr,"*** Paper empty pin active\n");
    if(pin_stat&BPP_SLCTIN_ERR)fprintf(stderr,"*** Select-in pin active\n");
    if(pin_stat&BPP_BUSY_ERR)  fprintf(stderr,"*** Busy pin active\n");
  }
  return 0;
}

int fill_buf(FILE *ifp, struct buffer *bp){

  int cgot;

  bp->start =  0;	/* Initialize buffer pointers. */
  bp->stop  =  0;       /* Remenber that [stop]-th elment is NOT valid. */

  while( bp->stop < BUFSIZE ){

    if( (cgot=getc(ifp)) == EOF){
      break;
    }
    else{
      bp->buf[bp->stop] = (char) cgot;
      (bp->stop)++;
    }
  }
  return ( bp->stop - bp->start );
}


/* Hardware reset */
int bpp_reset(int fd){
  extern int verbose;
  struct bpp_pins pins;

  ioctl(fd, BPPIOC_GETOUTPINS, &pins);

  pins.output_reg_pins |=  (u_char)BPP_INIT_PIN;
  ioctl(fd, BPPIOC_SETOUTPINS, &pins);

  usleep(RESET_DURATION);

  pins.output_reg_pins &= ~(u_char)BPP_INIT_PIN;
  ioctl(fd, BPPIOC_SETOUTPINS, &pins);

  if(verbose) fprintf(stderr, "*** Reset by hardware!\n");
  return 0;
}

/* Help message */
int bpp_help(){
  fprintf(stderr,"%s%s%s%s%s%s%s%s%s%s",
"  bppout --  output a file to ",BPPDEV,", the parallel interface.\n" ,
"  Usage:\n" ,
"    bppout [file]  : If `file' is missing, read the standard input.\n" ,
"    bppout -reset  : Reset the equipment at the interface by hardware.\n" ,
"    bppout -help   : Print this message.\n" ,
"    bppout -verb   : Print some additional information.\n" ,
"      Note: If the process is interrupted before completing data transfer,\n",
"            the equipment at the interface will be reset automatically.\n" );

  return 0;
}


/* Signal handling */
void sig_proc(){
  extern unsigned long cumcnt;
  extern int outfd;

  /* Do hardware reset if any data have been written. */
  if(cumcnt > 0) bpp_reset(outfd);
  exit(-1);

}
